// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include "MaFlash.h"

#include <getopt.h>

#include <iostream>
#include <fstream>
#include <algorithm>
#include <future>

#include <Logging.h>

#include <zip.h>

#include <openssl/rsa.h>
#include <openssl/pem.h>

using namespace std::literals;

namespace {

const char* s_pubKey =
    "-----BEGIN PUBLIC KEY-----\n"
    "MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAvawXFWKUX6GZ7SRh6E9M\n"
    "Rz8lP/f0YFmzoAc0IFcptKf/5pxbb14xc3TOAUrvAbbQ0/n8Y9ftArP1Aeez46QQ\n"
    "Rd689jmuXm0HfOUabQbeCEHEvmmMHGYA25J0l8YmkDZai+L0vK5uKj18uYNWrh0t\n"
    "v/5fpcw66OPeGOSS9Xkd64sM4CLfJCapI9tqDwUySgQDEt/B8+4MY9VlJZ2EfhAj\n"
    "+BC+HZwcEyHvR+2hl0YcSNl+KD1JJP7rYFmtC0N7Xr5lFXiwWuas7O8UF2iS8tdC\n"
    "SKfo4RRtlBl5WjqLjgJ6u1c0H9bNYFCd8uW9iB0MJE5KP0W9yBs/yFlGecrZTI5Q\n"
    "XQIDAQAB\n"
    "-----END PUBLIC KEY-----\n";

option s_options[] = {{"help", no_argument, nullptr, 'h'}, {"device", required_argument, nullptr, 'd'}, {"parallel", no_argument, nullptr, 'p'},
    {"stop-on-error", no_argument, nullptr, 's'}, {"backup", no_argument, nullptr, 'b'}, {nullptr, 0, nullptr, 0}};

void usage(const char* Program) {
  std::cout << "Usage: " << Program << " info <file>\n";
  std::cout << "       " << Program << " program {-d | --device} {<device>,... | all} [{-p | --parallel} | {-s | --stop-on-error}] [-b | --backup] <file> ]\n";
  std::cout << "           program the flash (primary or backup regions) with the contents of the specified file\n";
  std::cout << "       " << Program << " verify {-d | --device} {<device>,... | all} [{-p | --parallel} | {-s | --stop-on-error}] [-b | --backup] <file> ]\n";
  std::cout << "           verify the flash contents (primary or backup regions) with the contents of the specified file\n";
  std::cout << "\nOptions:\n\n";
  std::cout << "    -d | --device         a comma separated list of PCIe DBFS *or* the keyword \"all\" which will use all detected ma35 devices\n";
  std::cout << "    -p | --parallel       perform the program or verify operation simultaneous across all specified devices\n";
  std::cout << "    -s | --stop-on-error  for non-parallel operations, stop at the first error detected.  The default is to continue on error\n";
  std::cout << "    -b | --backup         specify that the program or verify operation should use the backup regions (where appropriate)\n";
}

#define OP(X) #X,
const char* s_regionNames[] = {FLASH_REGION_LIST};
const char* s_resultNames[] = {FLASH_RESULT_LIST};
#undef OP

void outputMeta(const nlohmann::ordered_json& Json) {
  bool first = true;
  for (const auto& [key, value] : Json.items()) {
    if (key == "signature") {
      continue;
    }
    if (!first) {
      std::cout << ", ";
    } else {
      first = false;
    }
    std::cout << key << ": " << std::string(value);
  }
}

template <typename F>
struct LambdaGuard {
  F m_f;
  explicit LambdaGuard(const F& Function) : m_f(Function) {
  }
  ~LambdaGuard() { std::invoke(m_f); }
};

} // namespace

auto MaFlash::ProcessContents(const std::string& File, std::string&& Contents) -> file_t {
  auto pos = Contents.find_first_of('\0');
  if (pos == std::string::npos) {
    std::cerr << "Unable to find header for " << File << std::endl;
    return FileResult(MISSING_HEADER);
  }
  json json;
  try {
    json = json.parse(Contents.substr(0, pos));
  } catch (...) {
    std::cerr << "Bad header for " << File << std::endl;
    return FileResult(BAD_HEADER);
  }
  Contents.erase(0, pos + 1);
  return {OK, File, json, std::move(Contents)};
}

auto MaFlash::GetNextZipFile() -> zipfile_t {
  if (m_zipIndex == m_zipNum) {
    return ZipFileResult(NO_MORE);
  }
  auto rawName = zip_get_name(m_zip, m_zipIndex, ZIP_FL_ENC_GUESS);
  if (!rawName) {
    std::cerr << "Error accessing container file: " << zip_error_strerror(zip_get_error(m_zip)) << '\n';
    return ZipFileResult(BAD_ZIP);
  }
  auto file = zip_fopen_index(m_zip, m_zipIndex, 0);
  if (!file) {
    std::cerr << "Error reading container file [" << rawName << "]: " << zip_error_strerror(zip_get_error(m_zip)) << '\n';
    return ZipFileResult(BAD_ZIP);
  }
  zip_stat_t stat;
  if (zip_stat_index(m_zip, m_zipIndex, 0, &stat)) {
    std::cerr << "Error getting size of container file [" << rawName << "]: " << zip_error_strerror(zip_get_error(m_zip)) << '\n';
    return ZipFileResult(BAD_ZIP);
  }
  ++m_zipIndex;
  return {OK, std::string(rawName), file, stat.size};
}

auto MaFlash::GetFile(const std::string& File) -> file_t {
  if (m_zip) {
    auto [result, name, zipfile, size] = GetNextZipFile();
    if (result == NO_MORE) {
      return FileResult(NO_MORE);
    }
    logDebug("ZIP %u %s %p", result, name.c_str(), zipfile);
    std::string contents;
    contents.reserve(size);
    contents.resize(size);
    if (zip_fread(zipfile, contents.data(), size) != size) {
      std::cerr << "Error reading container file [" << name << "]: " << zip_error_strerror(zip_file_get_error(zipfile)) << '\n';
      return FileResult(BAD_ZIP);
    }
    zip_fclose(zipfile);
    return ProcessContents(name, std::move(contents));
  }
  int error;
  m_zip = zip_open(File.c_str(), ZIP_RDONLY, &error);
  if (m_zip) {
    m_zipNum   = zip_uint64_t(zip_get_num_entries(m_zip, ZIP_FL_UNCHANGED));
    m_zipIndex = 0;
    return GetFile(File);
  }
  if (error != ZIP_ER_NOZIP) {
    std::cerr << "Error opening zip file: " << File << '\n';
    return FileResult(BAD_ZIP);
  }
  std::ifstream file(File);
  if (!file.is_open()) {
    std::cerr << "Cannot open: " << File << '\n';
    return FileResult(FILE_ERROR);
  }
  std::ostringstream buf;
  buf << file.rdbuf();
  if (buf.fail()) {
    std::cerr << "Error reading: " << File << '\n';
    return FileResult(FILE_ERROR);
  }
  auto  res    = ProcessContents(File, std::move(buf.str()));
  auto& result = std::get<0>(res);
  if (result == OK) {
    result = NO_MORE;
  }
  return res;
}

MaFlash::result_t MaFlash::DoInfo(const std::string& File) {
  while (true) {
    auto [result, name, json, contents] = GetFile(File);
    if (result != OK && result != NO_MORE) {
      return result;
    }
    if (name.empty()) {
      return OK;
    }
    logDebug("%s [%lu]", name.c_str(), contents.size());
    std::cout << name << ": ";
    outputMeta(json);
    std::cout << '\n';
    if (result == NO_MORE) {
      return OK;
    }
  }
}

auto MaFlash::ValidateDevices(const std::string& DeviceString) -> result_t {
  if (DeviceString == "all") {
    for (const auto& name : m_manager.getDeviceList()) {
      try {
        m_devices.emplace_back(m_manager.getDevice(name));
      } catch (std::exception& e) {
        std::cerr << "Error accessing device: " << name << " [" << e.what() << "]\n";
        return BAD_DEVICE;
      }
    }
    return OK;
  }
  std::stringstream        stream(DeviceString);
  auto                     deviceList = m_manager.getDeviceList();
  std::vector<std::string> seenDevices;
  std::string              val;
  while (std::getline(stream, val, ',')) {
    auto iter = std::ranges::find(deviceList, val);
    if (iter == deviceList.end()) {
      std::cerr << "Cannot locate device: " << val << '\n';
      return BAD_DEVICE;
    }
    if (std::ranges::find(seenDevices, val) != seenDevices.end()) {
      std::cerr << "Duplicate device: " << val << '\n';
      return BAD_DEVICE;
    }
    try {
      m_devices.emplace_back(m_manager.getDevice(val));
    } catch (std::exception& e) {
      std::cerr << "Error accessing device: " << val << " [" << e.what() << "]\n";
      return BAD_DEVICE;
    }
    seenDevices.emplace_back(std::move(val));
  }
  return OK;
}

bool MaFlash::VerifySignature(const json& Json, const std::string& Contents) {
  auto json     = Json;
  auto sigCoded = std::string(json["signature"]);
  json.erase("signature");
  uint8_t buf[256]; // known
  auto    bio = BIO_new_mem_buf(sigCoded.c_str(), -1);
  auto    b64 = BIO_new(BIO_f_base64());
  bio         = BIO_push(b64, bio);
  BIO_set_flags(bio, BIO_FLAGS_BASE64_NO_NL);
  auto length = BIO_read(bio, buf, sigCoded.size());
  if (length != 256) {
    return false;
  }
  auto keyio = BIO_new_mem_buf(s_pubKey, -1);
  auto key   = PEM_read_bio_RSA_PUBKEY(keyio, nullptr, nullptr, nullptr);
  BIO_free(keyio);
  auto pubKey = EVP_PKEY_new();
  EVP_PKEY_assign_RSA(pubKey, key);
  auto ctx   = EVP_MD_CTX_create();
  auto guard = LambdaGuard([=]() { EVP_MD_CTX_destroy(ctx); });
  if (EVP_DigestVerifyInit(ctx, nullptr, EVP_sha256(), nullptr, pubKey) <= 0) {
    return false;
  }
  auto header = std::string(json.dump(2));
  if (EVP_DigestVerifyUpdate(ctx, header.data(), header.size()) <= 0) {
    return false;
  }
  if (EVP_DigestVerifyUpdate(ctx, Contents.data(), Contents.size()) <= 0) {
    return false;
  }
  if (EVP_DigestVerifyFinal(ctx, buf, 256) != 1) {
    return false;
  }
  return true;
}

void MaFlash::Note(const std::string& Device, const std::string& Name, result_t Result) {
  m_notes << Device << " [" << std::left << std::setw(14) << s_resultNames[Result] << "] " << Name << '\n';
}

auto MaFlash::DoProgramVerify(bool Verify, const std::string& File, bool Parallel, bool StopOnError, bool Backup) -> result_t {
  result_t lastRes = OK;
  bool     noMore  = false;
  while (!noMore) {
    auto [result, name, json, contents] = GetFile(File);
    if (result != OK && result != NO_MORE) {
      return result;
    }
    if (name.empty()) {
      return OK;
    }
    noMore = (result == NO_MORE);
    if (!json.contains("type")) {
      std::cerr << "WARNING: " << name << " is missing type information, skipping\n";
      continue;
    }
    if (!json.contains("version")) {
      std::cerr << "WARNING: " << name << " is missing version information, skipping\n";
      continue;
    }
    if (!json.contains("signature")) {
      std::cerr << "WARNING: " << name << " is missing signature information, skipping\n";
      continue;
    }
    auto iter = std::ranges::find(s_regionNames, json["type"]);
    if (iter == std::end(s_regionNames)) {
      std::cerr << "WARNING: " << name << " has unsupported type [" << std::string(json["type"]) << "], skipping\n";
      continue;
    }
    std::string keyset;
    if (json.contains("keyset")) {
      keyset = json["keyset"];
      if (keyset != "AMD" && keyset != "VSI") {
        std::cerr << "WARNING: " << name << " has unsupported keyset [" << keyset << "], skipping\n";
        continue;
      }
    }
    auto type = region_t(std::distance(std::begin(s_regionNames), iter));
    if (Backup && type != ZSP && type != ESECURE && type != PCIE_PHY_PATCH) {
      std::cerr << "WARNING: Backup requested for file " << name << ", which has type [" << std::string(json["type"]) << "] with no backup slot, skipping\n";
      continue;
    }
    if (!VerifySignature(json, contents)) {
      std::cerr << "ERROR: Signature verification failed for file " << name << '\n';
      return BAD_SIG;
    }
    std::cout << "Using flash image: " << name << " [";
    outputMeta(json);
    std::cout << "]\n";
    std::vector<std::future<result_t>> futures;
    for (auto& device : m_devices) {
      result_t    res;
      std::string extra;
      if (!keyset.empty() && ((keyset == "AMD") != device->isAmd())) {
        std::cout << "  Skipping device " << device->Name() << " (mismatched keyset)\n";
        continue;
      }
      if (Parallel) {
        futures.emplace_back(std::async(std::launch::async, (Verify) ? &DeviceInterface::Verify : &DeviceInterface::Program, &*device, type, contents, Backup));
        continue;
      }
      std::cout << "  Device: " << device->Name() << '\n';
      try {
        res = (Verify) ? device->Verify(type, contents, Backup) : device->Program(type, contents, Backup);
      } catch (std::exception& e) {
        extra = e.what();
        res   = (Verify) ? VERIFY_FAILED : PROGRAM_FAILED;
      }
      Note(device->Name(), name, res);
      if (res != OK) {
        lastRes = res;
        std::cerr << "    **** ERROR " << ((Verify) ? "Verifying " : "Programming ") << name << " of type [" << std::string(json["type"]) << "] ";
        if (Backup) {
          std::cerr << "in backup region ";
        }
        std::cerr << extra << std::endl; // flush
        if (StopOnError) {
          noMore = true;
          break;
        }
      }
    }
    while (!futures.empty()) {
      result_t    res;
      std::string extra;
      std::cout << "  Device: " << m_devices[futures.size() - 1]->Name() << '\n';
      auto& future(futures.back());
      try {
        res = future.get();
      } catch (std::exception& e) {
        extra = e.what();
        res   = (Verify) ? VERIFY_FAILED : PROGRAM_FAILED;
      }
      futures.pop_back();
      if (res != OK) {
        lastRes = res;
        std::cerr << "    **** ERROR " << ((Verify) ? "Verifying" : "Programming ") << name << " of type [" << std::string(json["type"]) << "] ";
        if (Backup) {
          std::cerr << "in backup region ";
        }
        std::cerr << extra << std::endl; // flush
        if (StopOnError) {
          noMore = true;
          break;
        }
      }
    }
    if (result == NO_MORE) {
      return lastRes;
    }
  }
  return lastRes;
}

void MaFlash::DumpDeviceList() {
  auto list = m_manager.getDeviceList();
  if (list.empty()) {
    std::cerr << "\n*** No Devices Found ***\n";
    return;
  }
  std::cerr << "\nAvailable Devices:\n";
  for (const auto& name : list) {
    std::cerr << "  " << name << '\n';
  }
}

auto MaFlash::Run() -> result_t {
  logSetEnv();
  bool        backup      = false;
  bool        stopOnError = false;
  bool        parallel    = false;
  std::string deviceString;
  optind = 1; // allow reuse for testing
  while (true) {
    auto opt = getopt_long(m_argc, m_argv, "hd:psb", s_options, nullptr);
    if (opt == -1) {
      break;
    }
    switch (opt) {
    case 'b':
      backup = true;
      break;
    case 'p':
      parallel = true;
      break;
    case 's':
      stopOnError = true;
      break;
    case 'd':
      deviceString = optarg;
      break;
    case '?':
      std::cerr << "\nUnrecognized option: " << optarg << "\n\n";
      [[fallthrough]];
    case 'h':
      usage(m_argv[0]);
      return (opt != 'h') ? USAGE : OK;
    }
  }
  if (optind == m_argc) {
    std::cerr << "A command must be specified\n";
    return MISSING_COMMAND;
  }
  auto                                             command  = m_argv[optind++];
  static constexpr std::array<std::string_view, 3> commands = {"info"sv, "program"sv, "verify"sv};
  auto                                             iter     = std::ranges::find(commands, command);
  if (iter == std::end(commands)) {
    std::cerr << "No such command: " << command << '\n';
    return BAD_COMMAND;
  }
  auto cmdIndex = std::distance(std::begin(commands), iter);
  if (!cmdIndex && (!deviceString.empty() || parallel || stopOnError || backup)) {
    std::cerr << "Inapplicable options for info\n";
    return INAPPLICABLE_OPTIONS;
  }
  if (cmdIndex && stopOnError && parallel) {
    std::cerr << "--stop-on-error is not valid with --parallel\n";
    return INAPPLICABLE_OPTIONS;
  }
  if (cmdIndex && deviceString.empty()) {
    std::cerr << *iter << " requires --device\n";
    DumpDeviceList();
    return MISSING_DEVICE;
  }
  if (optind == m_argc) {
    std::cerr << "A file must be specified\n";
    return MISSING_FILE;
  }
  std::string file = m_argv[optind++];
  if (optind != m_argc) {
    std::cerr << "Extra arguments: " << m_argv[optind] << '\n';
    return EXTRA_ARGS;
  }
  if (!cmdIndex) {
    return DoInfo(file);
  }
  auto result = ValidateDevices(deviceString);
  if (result != OK) {
    DumpDeviceList();
    return result;
  }
  auto res = DoProgramVerify((cmdIndex == 2), file, parallel, stopOnError, backup);
  if (res != OK && !m_notes.str().empty()) {
    std::cerr << "\nErrors occurred:\n\n" << m_notes.str();
  }
  return res;
}
