// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

extern "C" {
#include <pci/pci.h>
}

#include <Logging.h>

#include <jsf_int.h>

using namespace std::literals;

namespace jsf_int {

std::unique_ptr<Jsf> CreateJsf() {
  return std::make_unique<Jsf>();
}

std::pair<Node*, std::string_view> Jsf::Find(std::string_view Path, bool Create) {
  NodeInternal* node = &m_root;
  if (Path.empty()) {
    return {node, {}};
  }
  while (true) {
    auto pos = Path.find_first_of('/');
    auto key = (pos == std::string_view::npos) ? Path : std::string_view(Path.begin(), Path.begin() + pos);
    if (auto iter = node->m_children.find(key); iter != node->m_children.end()) {
      if (pos == std::string_view::npos) {
        return {iter->get(), {}};
      } else {
        if (!iter->get()->m_isLeaf) {
          node = static_cast<NodeInternal*>(iter->get());
        } else {
          const auto leaf = static_cast<NodeLeaf*>(iter->get());
          if (leaf->IsPathCapable()) {
            Path.remove_prefix(pos + 1);
            return {iter->get(), Path};
          }
          return {};
        }
      }
    } else if (!Create) {
      return {};
    } else {
      auto val      = std::make_unique<NodeInternal>(std::string(key));
      val->m_parent = node;
      auto next     = val.get();
      node->m_children.emplace(std::move(val));
      node = next;
    }
    if (pos == std::string_view::npos) {
      return {node, {}};
    }
    Path.remove_prefix(pos + 1);
  }
  logInfo("FIND: %s", std::string(Path).c_str());
  return {};
}

Node* Jsf::Insert(Node* Parent, Node::ptr_t Node) {
  if (Parent->m_isLeaf) {
    return nullptr;
  }
  auto internal         = static_cast<NodeInternal*>(Parent);
  Node->m_parent        = Parent;
  auto [iter, inserted] = internal->m_children.emplace(std::move(Node));
  return (inserted) ? iter->get() : nullptr;
}

Node* Jsf::Insert(std::string_view Path, Node::ptr_t Node) {
  auto [node, path] = Find(Path, true);
  if (!node) {
    return nullptr;
  }
  logInfo("Node: %p", node);
  return Insert(node, std::move(Node));
}

json Jsf::RecurseBuild(Node* Node, const json& Params) {
  if (Node->m_isLeaf) {
    return static_cast<NodeLeaf*>(Node)->Invoke(Params, true);
  }
  json       j;
  const auto internal = static_cast<NodeInternal*>(Node);
  for (const auto& child : internal->m_children) {
    auto val = RecurseBuild(child.get(), Params);
    if (!val.is_null()) {
      j[child.get()->m_name] = val;
    }
  }
  return j;
}

std::string Jsf::Execute(const std::string& Request) {
  json resp;
  json req;
  try {
    req = json::parse(Request);
  } catch (std::exception& e) {
    resp["status"] = std::string("error parsing: ") + e.what();
    return resp.dump();
  }
  if (auto iter = req.find("path"); iter != req.end()) {
    auto [node, path] = Find(std::string(*iter));
    if (!node) {
      resp["status"] = "error: path not found";
      return resp.dump();
    }
    resp["status"] = "ok";
    json params;
    if (auto iter2 = req.find("parameters"); iter2 != req.end()) {
      params = *iter2;
    }
    if (!node->m_isLeaf) {
      resp["status"] = "ok";
      resp["data"]   = RecurseBuild(node, params);
      return resp.dump();
    }
    try {
      auto invoke = static_cast<jsf_int::NodeLeaf*>(node);
      if (invoke->IsPathCapable()) {
        resp["data"] = invoke->InvokeWithPath(path, params); // TODO(jkuhn): exception for error to set status?
      } else {
        resp["data"] = invoke->Invoke(params);
      }
    } catch (std::exception& e) {
      resp["status"] = std::string("ERROR: ") + e.what();
      return resp.dump();
    }
  } else {
    resp["status"] = "error: path must be specified";
  }
  return resp.dump();
}

}; // namespace jsf_int
