// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <jsf_helpers.h>

#include <thread>
#include <sys/utsname.h>
#include <sys/sysinfo.h>
#include <gnu/libc-version.h>

#include <Logging.h>

namespace jsf_helpers {

NodeVersionInfo::NodeVersionInfo(const std::string& Name, const std::string& VersionFile) : jsf_helpers::NodeRawFile(Name, VersionFile) {
}

json NodeVersionInfo::Invoke(const json& Parameters, bool Iteration) {
  std::string version;
  try {
    version = NodeRawFile::Invoke(Parameters, Iteration);
    if (!version.empty() && version.back() == '\n') {
      version.pop_back();
    }
  } catch (const std::runtime_error& err) {
    version = "N/A";
  }
  return version;
}

NodeSystemInfo::NodeSystemInfo(const std::string& Name) : jsf_int::NodeJson(Name, {}) {
  utsname os_info;
  uname(&os_info);
  if (uname(&os_info) < 0) {
    logFatal("uname error");
    abort();
  }
  m_value["os"]      = os_info.sysname;
  m_value["node"]    = os_info.nodename;
  m_value["release"] = os_info.release;
  m_value["version"] = os_info.version;
  m_value["machine"] = os_info.machine;

  const auto processor_count = std::thread::hardware_concurrency();
  m_value["cpu_cores"]       = processor_count;

  struct sysinfo sys_info;
  if (sysinfo(&sys_info) != 0) {
    logFatal("sysinfo error");
    abort();
  }
  m_value["memory_bytes"] = sys_info.totalram;
  m_value["glibc"]        = gnu_get_libc_version();
}

} // namespace jsf_helpers
