// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <sys/stat.h>
#include <jsf_helpers.h>
#include <thread>

#include "Flash.h"

#include <Logging.h>

volatile SPI_InitTypeDef spi_master_inst;

U32 Flash::TimerReadCountFunc(U32 TimerIx) {
  return m_device.ReadBar2(TIMER_CNT(TimerIx));
}

void Flash::Usleep(U32 Us) {
  U32 Start, End, Count;
  U32 Clk; //timer clock, unit:MHz
#if defined(FPGA_MODE) || defined(EMULATOR_MODE)
  Clk = 20; // 20MHz
#else
  Clk = 500; // 500 MHz
#endif
  if (0 == Us)
    return;
  Start = TimerReadCountFunc(0);
  while (TRUE) {
    End = TimerReadCountFunc(0);
    if (Start >= End)
      Count = Start - End;
    else
      Count = 0xFFFFFFFF - End + Start + 1;
    if ((Count / Clk) >= Us)
      break;
  }
}

void Flash::CommonDelayUsFunc(U32 Count) {
  Usleep(Count);
}

SPI_StatusTypeDef Flash::SPI_CheckStatus(U32 Reg, U32 Flag, U8 Level, U32 TimeOut) {
  U32 Count;
  Count = TimeOut;

  do {
    if (Level) {
      if ((m_device.ReadBar2(Reg)) & Flag) {
        return SPI_OK;
      }
    } else {
      if (((m_device.ReadBar2(Reg)) & Flag) == 0)
        return SPI_OK;
    }
    Count--;
  } while ((Count != 0) || (TimeOut == 0));
  return SPI_TIMEOUT;
}

void Flash::SPIChipDeselectFunc(void) {
  m_device.WriteBar2<uint8_t>(SPI_CS(spi_master_inst.instance), 0x0); //set cs to high
}

SPI_StatusTypeDef Flash::SPI_Receive(SPI_InitTypeDef* spi_instance, U8* pRxData, U32 Len) {
  U32 val32 = 0;
  U32 ix    = 0;
  U32 rxLen, tmp_readLen;
  U32 rxtrig_level = 0;
  U32 remain       = Len;
  U8* temp_buff    = pRxData;

  rxtrig_level = m_device.ReadBar2(SPI_FIFO_RXTRIG_LVL(spi_instance->instance));
  val32        = m_device.ReadBar2(SPI_CTL3(spi_instance->instance));
  m_device.WriteBar2<uint8_t>(SPI_CTL3(spi_instance->instance), val32 | BATCH_DISABLE);

  do {
    rxLen  = remain;
    remain = remain - rxLen;

    val32 = m_device.ReadBar2(SPI_CTL1(spi_instance->instance));
    m_device.WriteBar2<uint8_t>(SPI_CTL1(spi_instance->instance), val32 | RX_ENABLE);

    while (rxLen > 0) {
      if (rxLen > rxtrig_level) {
        for (ix = 0; ix < rxtrig_level; ix++) {
          m_device.WriteBar2<uint32_t>(SPI_TXREG(spi_instance->instance), 0);
        }
        if (SPI_CheckStatus(SPI_STATUS1(spi_instance->instance), RX_ALMOST_FULL, TRUE, spi_instance->timeout) == SPI_OK) {
          tmp_readLen = rxtrig_level;
          for (ix = 0; ix < tmp_readLen; ix++) {
            unsigned int val = m_device.ReadBar2(SPI_RXREG(spi_instance->instance));
            temp_buff[ix]    = val;
          }
          rxLen -= tmp_readLen;
          temp_buff += tmp_readLen;
        } else {
          return SPI_ERROR;
        }
      } else {
        tmp_readLen = rxLen;
        for (ix = 0; ix < tmp_readLen; ix++) {
          m_device.WriteBar2<uint32_t>(SPI_TXREG(spi_instance->instance), 0);
        }
        for (ix = 0; ix < tmp_readLen; ix++) {
          if (SPI_CheckStatus(SPI_STATUS2(spi_instance->instance), RXFIFO_EMPTY, FALSE, spi_instance->timeout) == SPI_OK) {
            unsigned int val = m_device.ReadBar2(SPI_RXREG(spi_instance->instance));
            temp_buff[ix]    = val;
          } else {
            return SPI_ERROR;
          }
        }
        rxLen -= tmp_readLen;
        temp_buff += tmp_readLen;
      }
    }
    if (SPI_CheckStatus(SPI_STATUS2(spi_instance->instance), TXFIFO_EMPTY, TRUE, spi_instance->timeout) == SPI_TIMEOUT) {
      return SPI_ERROR;
    }
    if (SPI_CheckStatus(SPI_STATUS2(spi_instance->instance), RXFIFO_EMPTY, TRUE, spi_instance->timeout) == SPI_TIMEOUT) {
      return SPI_ERROR;
    }
    val32 = m_device.ReadBar2(SPI_CTL1(spi_instance->instance));
    m_device.WriteBar2<uint8_t>(SPI_CTL1(spi_instance->instance), val32 & (~RX_ENABLE));
  } while (remain != 0);

  return SPI_DONE;
}

void Flash::SPIMasterReceiveFunc(U8* Buff, U32 RxLen) {
  if (SPI_Receive((SPI_InitTypeDef*) ((void*) &spi_master_inst), Buff, RxLen) == SPI_ERROR) {
    throw std::runtime_error("SPI_Receive timeout!");
  }
}

SPI_StatusTypeDef Flash::SPI_Transmit(SPI_InitTypeDef* spi_instance, U8* pTxData, U32 Len) {
  U32 val32 = 0;
  U32 ix    = 0;
  U32 txLen, tmp_writeLen;
  U32 txtrig_level = 0;
  U32 remain       = Len;
  U8* temp_buff    = pTxData;

  txtrig_level = m_device.ReadBar2(SPI_FIFO_TXTRIG_LVL(spi_instance->instance));
  val32        = m_device.ReadBar2(SPI_CTL3(spi_instance->instance));
  m_device.WriteBar2<uint8_t>(SPI_CTL3(spi_instance->instance), val32 | BATCH_DISABLE);

  do {
    txLen        = remain;
    remain       = remain - txLen;
    tmp_writeLen = MIN(txLen, FIFO_DEPTH); // first write fifo depth data to tx reg
    for (ix = 0; ix < tmp_writeLen; ix++) {
      m_device.WriteBar2(SPI_TXREG(spi_instance->instance), temp_buff[ix]);
    }
    txLen -= tmp_writeLen;
    temp_buff += tmp_writeLen;
    val32 = m_device.ReadBar2(SPI_CTL1(spi_instance->instance));
    m_device.WriteBar2<uint8_t>(SPI_CTL1(spi_instance->instance), val32 | TX_ENABLE); //enable tx

    while (txLen > 0) {
      if (SPI_CheckStatus(SPI_STATUS1(spi_instance->instance), TX_ALMOST_EMPTY, TRUE, spi_instance->timeout) == SPI_OK) {
        tmp_writeLen = MIN(txLen, FIFO_DEPTH - txtrig_level);
        for (ix = 0; ix < tmp_writeLen; ix++) {
          m_device.WriteBar2(SPI_TXREG(spi_instance->instance), temp_buff[ix]);
        }
        txLen -= tmp_writeLen;
        temp_buff += tmp_writeLen;
      } else
        return SPI_ERROR;
    }

    if (SPI_CheckStatus(SPI_STATUS2(spi_instance->instance), TXFIFO_EMPTY, TRUE, spi_instance->timeout) == SPI_TIMEOUT)
      return SPI_ERROR;

    val32 = m_device.ReadBar2(SPI_CTL1(spi_instance->instance));
    m_device.WriteBar2<uint8_t>(SPI_CTL1(spi_instance->instance), val32 & (~TX_ENABLE));
  } while (remain != 0);

  return SPI_DONE;
}

void Flash::SPIMasterTransmitFunc(U8* Buff, U32 TxLen) {
  if (SPI_Transmit((SPI_InitTypeDef*) ((void*) &spi_master_inst), Buff, TxLen) == SPI_ERROR) {
    throw std::runtime_error("SPI_Transmit timeout!");
  }
}

void Flash::SPIChipSelectFunc(void) {
  m_device.WriteBar2<uint8_t>(SPI_CS(spi_master_inst.instance), (spi_master_inst.cs + 0x0));
}

void Flash::SPIFlashReadFunc(U32 Addr, U8* pData, U32 Count) {
  U8  Cmd[4];
  U8* BuffTemp          = 0;
  U32 ReadLen           = 0;
  U32 AddrTemp          = Addr;
  U32 Remain            = Count;
  U32 FirstPageRWRemain = (FLS_PAGE_SIZE - Addr % FLS_PAGE_SIZE) % FLS_PAGE_SIZE;
  BuffTemp              = pData;

  do {
    if (FirstPageRWRemain == 0)
      ReadLen = MIN(Remain, FLS_SECTOR_SIZE);
    else {
      ReadLen           = MIN(Remain, FirstPageRWRemain);
      FirstPageRWRemain = 0;
    }
    // Select the device
    SPIChipSelectFunc();

    Cmd[0] = FLS_CMD_READ; // Send the read command.

    // Send the address of the first byte to read.
    Cmd[1] = ((AddrTemp >> 16) & 0xFF);
    Cmd[2] = ((AddrTemp >> 8) & 0xFF);
    Cmd[3] = ((AddrTemp >> 0) & 0xFF);

    SPIMasterTransmitFunc(Cmd, 4);
    SPIMasterReceiveFunc(BuffTemp, ReadLen); // Read the next data byte from the receive FIFO and place it into the data buffer.
    SPIChipDeselectFunc();
    CommonDelayUsFunc(50);
    Remain = Remain - ReadLen;
    AddrTemp += ReadLen;
    BuffTemp += ReadLen;
  } while (Remain != 0);
}

void Flash::SPI_Init(SPI_InitTypeDef* SPI_InitType) {
  U32   val32;
  float divisorScale;
  U32   divisor   = 0;
  U32   prescaler = 0;
  if (SPI_InitType->ahbBusFreq < (2 * SPI_InitType->sclk))
    SPI_InitType->sclk = SPI_InitType->ahbBusFreq / 2;

  divisorScale = (SPI_InitType->ahbBusFreq / SPI_InitType->sclk) / 32;
  while (divisorScale >= 1.0) {
    divisor++;
    divisorScale = divisorScale / 2;
  }
  prescaler = ((SPI_InitType->ahbBusFreq / SPI_InitType->sclk) / (2 << divisor)) - 1;
  val32     = (U32) (SPI_InitType->ahbBusFreq / SCLK_DIV(divisor, prescaler));

  m_device.WriteBar2<uint8_t>(SPI_SCLK_CON(SPI_InitType->instance), SCLK_VAL(divisor, prescaler));

  //set SPI work mode
  val32 = (U32) ((SPI_InitType->mstSlvMode) | (SPI_InitType->workMode) | (SPI_InitType->fifoWidth) | (SPI_InitType->firstBit));
  m_device.WriteBar2<uint8_t>(SPI_CTL1(SPI_InitType->instance), val32);

  //set SPI trig level
  m_device.WriteBar2<uint8_t>(SPI_FIFO_RXTRIG_LVL(SPI_InitType->instance), SPI_InitType->rxTrigLevel);
  m_device.WriteBar2<uint8_t>(SPI_FIFO_TXTRIG_LVL(SPI_InitType->instance), SPI_InitType->txTrigLevel);
  m_device.WriteBar2<uint8_t>(SPI_CTL2(SPI_InitType->instance), 0);
  m_device.WriteBar2<uint8_t>(SPI_CS(SPI_InitType->instance), 0x00);

  //software reset SPI FIFO
  m_device.WriteBar2<uint8_t>(SPI_STATUS1(SPI_InitType->instance), MODF_CLR | BATCH_RXDONE | BATCH_TXDONE);
  m_device.WriteBar2<uint8_t>(SPI_CTL3(SPI_InitType->instance), FIFO_RESET);
  m_device.WriteBar2<uint8_t>(SPI_CTL3(SPI_InitType->instance), 0x00);
}

void Flash::SPIMasterInitFunc(U8 SPINum, U32 IpClk) {
  if (spi_master_inst.inited == 0x55)
    return;

  memset((void*) &spi_master_inst, 0x0, sizeof(spi_master_inst));
  spi_master_inst.ahbBusFreq = IpClk;

  spi_master_inst.sclk        = SPI_BAUDRATE; // BaudRate;
  spi_master_inst.fifoWidth   = FIFO_WIDTH8;
  spi_master_inst.firstBit    = MSB;
  spi_master_inst.workMode    = SPI_MODE0;
  spi_master_inst.mstSlvMode  = MST_MODE;
  spi_master_inst.txTrigLevel = FIFO_DEPTH / 2;
  spi_master_inst.rxTrigLevel = FIFO_DEPTH / 2; // (FIFO_DEPTH/2 from ZSP FW code)
  spi_master_inst.intCfg      = 0;
  spi_master_inst.timeout     = TIMEOUT_1S;
  spi_master_inst.cs          = SPI_CS0;
  spi_master_inst.instance    = SPINum;

  SPI_Init((SPI_InitTypeDef*) &spi_master_inst);
  spi_master_inst.inited = 0x55;
}

void Flash::DumpFlashBuffer(U8* fbuff, U32 len, std::string Read_Flash_File) {
  remove(Read_Flash_File.c_str());
  std::fstream sf;
  sf.open(Read_Flash_File, std::ios::app | std::ios::binary);
  if (!sf) {
    logNotice("Flash file: %s -- not opened for Reading !", Read_Flash_File.c_str());
    m_device.fail();
  }
  U8* flash_buff_ptr = fbuff;
  for (unsigned int i = 0; i < len; i++) {

#if 0
    printf("%02x ", flash_buff_ptr[i]);
    if((i+1)%16 == 0)
      printf("\n");
#endif
    sf.write((char*) &flash_buff_ptr[i], sizeof(flash_buff_ptr[i]));
  }
  flash_buff_ptr = NULL;
  sf.close();
}

// READ
void Flash::ReadFlash(U32 Addr, U32 Len, const std::string& Read_Dump_Flash_Mem_File) {
  std::unique_ptr<U8[]> flash_buff;
  try {
    flash_buff = std::make_unique<U8[]>(SPI_FLASH_BUFFERSIZE);
  } catch (std::exception& e) {
    throw std::runtime_error("Could not allocate memory for flash_buff reading.");
  }

  SPIFlashReadFunc(Addr, flash_buff.get(), Len);
  DumpFlashBuffer(flash_buff.get(), Len, Read_Dump_Flash_Mem_File);
}

void Flash::ReadVerifyFlash(U32 Addr, U32 Len, U8* pBuff) {
  SPIFlashReadFunc(Addr, pBuff, Len);
}

// ERASE
void Flash::SPIFlash_Write_Enable(void) {
  U8 cmd[1];

  SPIChipSelectFunc();   // Chip select
  cmd[0] = FLS_CMD_WREN; // Send the write enable command
  SPIMasterTransmitFunc(cmd, 0x01);
  SPIChipDeselectFunc(); // Deselect
  CommonDelayUsFunc(50);
}

void Flash::SPIFlash_Write_Disable(void) {
  U8 cmd[1];
  // Chip select
  SPIChipSelectFunc();
  // Send the write disable command
  cmd[0] = FLS_CMD_WRDIS;
  SPIMasterTransmitFunc(cmd, 0x01);
  SPIChipDeselectFunc(); // Deselect
  CommonDelayUsFunc(50);
}

int Flash::SPIFlash_Checkbusy(void) {
  // Send the write status register command.
  U8  cmd[1];
  U32 timeout = TIMEOUT_1S;
  while (timeout--) {
    SPIChipSelectFunc();   //select the device
    cmd[0] = FLS_CMD_RSRL; //0x05 // send command
    SPIMasterTransmitFunc(cmd, 1);
    SPIMasterReceiveFunc(cmd, 1);

    SPIChipDeselectFunc();      // deselect device
    if (cmd[0] & FLS_STS_WIP) { // Check the busy bit
      // CommonDelayUsFunc(50);
      usleep(2);
    } else
      break;
  }

  if (0 == timeout) {
    return -ETIMEDOUT;
  } else {
    return 0;
  }
}

void Flash::SPIFlash_Disable_Protect(void) {
  U8 cmd[3];

  SPIChipSelectFunc(); // chip select
  cmd[0] = FLS_CMD_WRSR;
  cmd[1] = 0x00; // disable all protect
  cmd[2] = 0x00; // disable all protect
  SPIMasterTransmitFunc(cmd, 0x02);
  SPIChipDeselectFunc(); // deselect
}

void Flash::SPIFlashSectorEraseFunc(U32 addr) {
  U8 cmd[4];

  SPIFlash_Write_Enable();         // Enable flash write
  SPIFlash_Disable_Protect();      // Disable all block protect
  if (SPIFlash_Checkbusy() != 0) { // Reads the SPI flash status register.
    throw std::runtime_error("SPIFlash_Read_Status timeout!");
  }
  SPIFlash_Write_Enable(); // Enables SPI flash write operations
  SPIChipSelectFunc();     // Chip select

  addr   = addr & 0xfffff000;
  cmd[0] = FLS_CMD_SE;
  cmd[1] = ((addr >> 16) & 0xFF);
  cmd[2] = ((addr >> 8) & 0xFF);
  cmd[3] = ((addr >> 0) & 0xFF);
  SPIMasterTransmitFunc(cmd, 0x04);
  SPIChipDeselectFunc(); // Chip deselect
  CommonDelayUsFunc(50);

  if (SPIFlash_Checkbusy() != 0) { // SPIFlash_Checkbusy() must be used to query the SPI flash to determine when the sector erase operation has completed.
    throw std::runtime_error("Sector erase operation NOT completed (timeout)!");
  }
  std::this_thread::sleep_for(std::chrono::milliseconds(100));
}

void Flash::SPIFlashEraseFunc(U32 flash_addr, U32 len) {
  U32 sectors = 0;
  U32 i       = 0;

  sectors = (len + FLS_SECTOR_SIZE - 1) / FLS_SECTOR_SIZE;

  logNotice("EraseFlash Started..");
  U32 tmp_sectors = sectors / 10;

  for (i = 0; i < sectors; i++) {
    SPIFlashSectorEraseFunc(flash_addr + i * FLS_SECTOR_SIZE);
    if (i && i % tmp_sectors == 0) {
      logNotice("erase_progress: %d%%", (i * 100) / sectors);
    }
  }
}

void Flash::EraseFlash(U32 Addr, U32 Len) {
  SPIFlashEraseFunc(Addr, Len);
}

// WRITE
void Flash::SPIFlashPageProgramFunc(U8* src_buff, U32 dest_flash_addr, U32 len) {
  U8  cmd[4];
  U32 write_len;
  U8* buff_temp;
  U32 dest_flash_addr_temp = dest_flash_addr;
  U32 remain               = len;
  U32 first_page_rw_remain = (FLS_PAGE_SIZE - dest_flash_addr % FLS_PAGE_SIZE) % FLS_PAGE_SIZE;
  buff_temp                = src_buff;

  U32 flash_progress             = 0;
  U32 flash_progress_status      = 0;
  U32 prev_flash_progress_status = 0;

  logNotice("WriteFlash Starting, please Wait....");
  do {
    SPIFlash_Write_Enable(); // enable write
    if (first_page_rw_remain == 0)
      write_len = MIN(remain, FLS_MAX_WRITE_SIZE); //MIN(remain,FIFO_DEPTH);
    else {
      write_len            = MIN(remain, first_page_rw_remain);
      first_page_rw_remain = 0;
    }

    flash_progress += write_len;
    flash_progress_status = (flash_progress * 100) / len;

    if ((prev_flash_progress_status != flash_progress_status) && (flash_progress_status % 10 == 0)) {
      prev_flash_progress_status = flash_progress_status;
      logNotice("flash_progress = %d%%", flash_progress_status);
    }

    //
    SPIChipSelectFunc(); //chip select

#if 0
    // Debug SPIFlashPageProgramFunc
    logNotice("Debugging Write (SPIFlashPageProgramFunc): dest_flash_addr_temp=0x%08x ; write_len=%d", dest_flash_addr_temp, write_len);
    U8* tptr = buff_temp;
    for(unsigned int k=0; k<write_len; ++k) {
      printf("%02x ", tptr[k]);
      if((k+1)%16 == 0) {
        printf("\n");
      }
    }
#endif

    cmd[0] = FLS_CMD_PP;
    cmd[1] = (dest_flash_addr_temp >> 16) & 0xFF;
    cmd[2] = (dest_flash_addr_temp >> 8) & 0xFF;
    cmd[3] = (dest_flash_addr_temp >> 0) & 0xFF;
    SPIMasterTransmitFunc(cmd, 0x04);
    SPIMasterTransmitFunc(buff_temp, write_len);
    SPIChipDeselectFunc(); // deselect

    if (SPIFlash_Checkbusy() != 0) {
      throw std::runtime_error("SPIFlash_Read_Status timeout!");
    }
    std::this_thread::sleep_for(std::chrono::milliseconds(30));
    remain = remain - write_len;
    dest_flash_addr_temp += write_len;
    buff_temp += write_len;
  } while (remain != 0);

  SPIFlash_Write_Disable();
}

void Flash::WriteFlash(const char* Fname, U32 Addr, U32 Len) {
  U8* buff = (U8*) calloc(1, SPI_FLASH_BUFFERSIZE);
  if (!buff) {
    logNotice("Could not allocate memory for buff for Writing Flash File !");
    m_device.fail();
  }

  // Read File to be Programmed
  std::ifstream ffile;
  ffile.open(Fname, std::ios::in | std::ios::binary);
  if (!ffile) {
    logNotice("ZSP FW file Not opened for Reading !");
    m_device.fail();
  }
  ffile.read((char*) buff, SPI_FLASH_BUFFERSIZE);
#if 0
  //Debug
  logNotice("Debugging Write:");
  U8* tptr = buff;
  for(unsigned int k=0; k<Len; k++) {
    printf("%02x ", tptr[k]);
    if((k+1)%16 == 0)
      printf("\n");
  }
  printf("\n");
#endif

  SPIFlashPageProgramFunc((U8*) (buff), Addr, Len);
  free(buff);
  buff = NULL;
}

int Flash::VerifyFlashedRegion(U32 mem_addr, U32 mem_len, std::string flash_file) { // Read SPI region to buffer and compare with Golden file
  std::unique_ptr<U8[]> flash_buff;
  try {
    flash_buff = std::make_unique<U8[]>(mem_len);
  } catch (std::exception& e) {
    throw std::runtime_error("Could NOT allocate memory for flash_buff reading.");
  }
  ReadVerifyFlash(mem_addr, mem_len, flash_buff.get()); // Read SPI Region

  // Read Golden file to memory and Compare Buffers (Golden vs Current SPI Buff)
  auto          spi_buff = std::make_unique<char[]>(mem_len);
  std::ifstream fin(flash_file, std::ios::in | std::ios::binary);
  fin.read(spi_buff.get(), mem_len);
  int cmp_stat = memcmp(spi_buff.get(), flash_buff.get(), mem_len);

  return cmp_stat;
}

json Flash::Invoke(const json& Parameters, bool Iteration) {
  if (Iteration) {
    return {};
  }

  if (!m_device.IsDemo()) {
    SPIMasterInitFunc(0x0, SPI_AHB_CLK);
  }

  json        flashInfo = {};
  uint32_t    mem_addr  = 0;
  uint32_t    mem_len   = 0;
  std::string flash_file_type;

  if (Parameters.contains("action")) {
    if (Parameters.contains("file_type")) { // flashFileType (ZSP_A, ZSP_B, eSecure_A, eSecure_B, etc.)
      flash_file_type = Parameters["file_type"];
      if (flash_file_type == "ZSP") {
        mem_addr = ZSP_A_ADDR;
        mem_len  = ZSP_SIZE;
      } else if (flash_file_type == "ZSP_B") {
        mem_addr = ZSP_B_ADDR;
        mem_len  = ZSP_SIZE;
      } else if (flash_file_type == "eSecure") {
        mem_addr = ESecure_A_ADDR;
        mem_len  = ESecure_SIZE;
      } else if (flash_file_type == "eSecure_B") {
        mem_addr = ESecure_B_ADDR;
        mem_len  = ESecure_SIZE;
      } else if (flash_file_type == "PcieCtlPatch") {
        mem_addr = PcieCtlPatch_ADDR;
        mem_len  = PcieCtlPatch_SIZE;
      } else if (flash_file_type == "PciePhyPatch") {
        mem_addr = PciePhyPatch_A_ADDR;
        mem_len  = PciePhyPatch_SIZE;
      } else if (flash_file_type == "PciePhyPatch_B") {
        mem_addr = PciePhyPatch_B_ADDR;
        mem_len  = PciePhyPatch_SIZE;
      } else if (flash_file_type == "PcieFw") {
        mem_addr = PcieFw_ADDR;
        mem_len  = PcieFw_SIZE;
      } else if (flash_file_type == "spi_flash") {
        mem_addr = SPI_FLASH_ADDR;
        mem_len  = SPI_FLASH_SIZE;
      } else if (flash_file_type == "SC") {
        mem_addr = SC_ADDR;
        mem_len  = SC_SIZE;
      } else if (flash_file_type == "Reserved") {
        mem_addr = Reserved_ADDR;
        mem_len  = Reserved_SIZE;
      } else {
        throw std::runtime_error("Incorrect flash file type");
      }
    } else {
      throw std::runtime_error("file_type (flashFileType) required");
    }

    // Programming to Flash Memory, Verifying given Flash file
    if ((Parameters["action"] == "program" || Parameters["action"] == "flash_verify") && mem_len != 0) { // '"action": "program", "file": "flash"'
      if (Parameters.contains("file")) {
        struct stat sb_w;
        std::string flash_file = Parameters["file"]; // flash_file.erase(remove(flash_file.begin(), flash_file.end(), '\"'), flash_file.end()); // remove quotes
        if (stat(flash_file.c_str(), &sb_w) == -1) {
          throw std::runtime_error("Flash file path error");
        }

        // Check File-Size
        mem_len = sb_w.st_size;
        if (flash_file_type == "ZSP" && mem_len > ZSP_SIZE) {
          throw std::runtime_error("Size of ZSP FW file is greater than max Size");
        } else if (flash_file_type == "eSecure" && mem_len > ESecure_SIZE) {
          throw std::runtime_error("Size of eSecure FW file is greater than max Size");
        } else if (flash_file_type == "spi_flash" && mem_len > SPI_FLASH_SIZE) {
          throw std::runtime_error("Size of spi_flash file is greater than max Size");
        } else if (flash_file_type == "SC" && mem_len > SC_SIZE) {
          throw std::runtime_error("Size of SC FW file is greater than max Size");
        }

#if 0
        if (mem_addr == 0) {
          logNotice("Writing Full image is not supported");
          return {};
        }
#endif

        std::string       dev_bdf = m_device.GetBdf();
        std::stringstream faddr;
        faddr << "0x" << std::hex << mem_addr;
        faddr << ":0x" << std::hex << mem_len;

        if (m_device.IsDemo()) {
          return "PASS";
        }

        if (Parameters["action"] == "program") { // Call Flash Program Routine

          // TODO - Disabling this due to compatibility issues. Enable this in post alpha release.
          // auto flash_status = m_device.ReadBar4(m_device.FLASH_STATUS); // flash_status
          // if (flash_status) {
          //   throw std::runtime_error("SC is accessing flash, pls try FW update after sometime.");
          // }

          if (mem_addr == SC_ADDR) {
            auto asic_num = m_device.ReadBar4(m_device.ASIC_NUM);
            if (asic_num) {
              throw std::runtime_error("SC update is not supported from ASIC 1. Please try with ASIC 0.");
            }
            auto sc_update_status = m_device.ReadBar4(m_device.SC_UPDATE_STATUS); // sc_update_status
            if (sc_update_status & 0x0101) {
              throw std::runtime_error("SC is busy.");
            }
          }

          try {
            EraseFlash(mem_addr, mem_len);
            WriteFlash(flash_file.c_str(), mem_addr, mem_len);
          } catch (const std::exception& e) {
            logNotice("--> Reboot machine and try Flash-Program again.");
            throw;
          }
        }

        // Verify Flashed Region

#if 0
        ReadFlash(mem_addr, mem_len, "/tmp/vfio_dump_spi_flash_buff.bin");
        bool flash_status = compare_flashed_files(flash_file, "/tmp/vfio_dump_spi_flash_buff.bin");  // Read SPI region to file and compare with Golden file
#endif
        bool flash_status = 1;
        try {
          flash_status = VerifyFlashedRegion(mem_addr, mem_len, flash_file); // 0 => contents of both memory blocks are equal
        } catch (const std::exception& e) {
          logNotice("--> Reboot machine and try Flash program/verify again.");
          throw;
        }

        if (!flash_status) {

          if (mem_addr == SC_ADDR) {
            logNotice("SC successfully transferrred to Spi flash at addr: 0x%08x", mem_addr);
            m_device.WriteBar4<uint32_t>(m_device.SC_UPDATE_FLAG, 1); // sc_update_flag=1

            auto ib_sc_timout      = std::chrono::system_clock::now() + std::chrono::minutes{2}; // 2min;
            bool sc_update_success = 0;

            while (std::chrono::system_clock::now() < ib_sc_timout) {
              std::this_thread::sleep_for(std::chrono::seconds(5));
              auto sc_update_status = m_device.ReadBar4(m_device.SC_UPDATE_STATUS); // sc_update_status
              if (!sc_update_status) {
                sc_update_success = 1;
                logNotice("SC successfully updated.");
                break;
              } else if (sc_update_status) {
                logNotice("SC update is in progress..");
                continue;
              } else {
                throw std::runtime_error("SC update failed!");
              }
            }
            if (!sc_update_success) {
              throw std::runtime_error("SC update failed!");
            }
          }
          return "PASS";
        } else {
          if (Parameters["action"] == "program") {
            throw std::runtime_error("Flash program failed.");
          } else {
            throw std::runtime_error("Flash verification failed.");
          }
        }

        return flashInfo;
      }
      throw std::runtime_error("File (flashFile) required");
    }

    if (Parameters["action"] == "read" && mem_len != 0) {
      if (Parameters.contains("file")) {
        std::string read_dump_flash_mem = Parameters["file"];

        std::string       dev_bdf = m_device.GetBdf();
        std::stringstream faddr;
        faddr << "0x" << std::hex << mem_addr;
        faddr << ":0x" << std::hex << mem_len;

        if (m_device.IsDemo()) {
          return "PASS";
        }

        try {
          ReadFlash(mem_addr, mem_len, read_dump_flash_mem);
        } catch (const std::exception& e) {
          logNotice("--> Reboot machine and try Flash-Read again.");
          throw;
        }
        return "PASS";
      }
      throw std::runtime_error("File (read_flash_mem) required");
    }
  }
  if (!Parameters.is_null()) {
    throw std::runtime_error("Action required / invalid action");
  }
  return {};
}
